/*
  (c) 2005 Antonio Cunei, All rights reserved.

  This file is part of the Runabout Emulation Layer for PolyD.

  The Runabout Emulation Layer for PolyD is free software;
  you can redistribute it and/or modify it under the terms of
  the GNU General Public License as published by the Free
  Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  The Runabout Emulation Layer for PolyD is distributed
  in the hope that it will be useful, but WITHOUT ANY WARRANTY;
  without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE. See the GNU General Public License
  for more details.

  You should have received a copy of the GNU General Public License
  along with the Runabout Emulation Layer for PolyD; if not, write
  to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
  Boston, MA  02111-1307  USA
*/

package ovm.polyd.runemu;

import java.util.AbstractCollection;
import java.util.HashSet;

import ovm.polyd.policy.*;
import ovm.polyd.Descriptor;
import ovm.polyd.PolyD;

/**
 * An accurate Runabout emulation, this class can be used as
 * a "drop-in" replacement.<p>All major features are supported,
 * including support for primitives and the <code>visitDefault</code>
 * facility. It leaves out only the calls
 * that are more closely associated with the internal
 * implementation of the original Runabout:
 * <code>addExternalVisit</code> and <code>getAppropriateCode</code>.
 * 
 * @version $Revision: 1.2 $, $Date: 2005/04/02 06:03:47 $
 * @author Antonio Cunei
 *
 */
public class RunaboutBis {

 protected void err(Class c) {
  throw new RunaboutException(
   "The Runabout compatibility layer of PolyD has detected\n"+
   "a visit() call without a proper handler,\n"+
   "when dispatching using "+this.getClass()+",\n"+
   "calling visit("+c.getName()+")"
  );
 }
 protected void visitDefault(Object o)  { err(o.getClass()); }
 protected void visitDefault(char o)    { err(char.class); }
 protected void visitDefault(byte o)    { err(byte.class); }
 protected void visitDefault(int o)     { err(int.class); }
 protected void visitDefault(long o)    { err(long.class); }
 protected void visitDefault(float o)   { err(float.class); }
 protected void visitDefault(double o)  { err(double.class); }
 protected void visitDefault(short o)   { err(short.class); }
 protected void visitDefault(boolean o) { err(boolean.class); }

 /**
  * A "catch-all" <code>visit</code> method, redirected to
  * <code>visitDefault</code>.
  * @param o the object that should be visited
  */
 public void visit(Object o)  { visitDefault(o); }
 /**
  * A "catch-all" <code>visit</code> method, redirected to
  * <code>visitDefault</code>.
  * @param o the <code>char</code> that should be visited
  */
 public void visit(char o)    { visitDefault(o); }
 /**
  * A "catch-all" <code>visit</code> method, redirected to
  * <code>visitDefault</code>.
  * @param o the <code>byte</code> that should be visited
  */
 public void visit(byte o)    { visitDefault(o); }
 /**
  * A "catch-all" <code>visit</code> method, redirected to
  * <code>visitDefault</code>.
  * @param o the <code>int</code> that should be visited
  */
 public void visit(int o)     { visitDefault(o); }
 /**
  * A "catch-all" <code>visit</code> method, redirected to
  * <code>visitDefault</code>.
  * @param o the <code>long</code> that should be visited
  */
 public void visit(long o)    { visitDefault(o); }
 /**
  * A "catch-all" <code>visit</code> method, redirected to
  * <code>visitDefault</code>.
  * @param o the <code>float</code> that should be visited
  */
 public void visit(float o)   { visitDefault(o); }
 /**
  * A "catch-all" <code>visit</code> method, redirected to
  * <code>visitDefault</code>.
  * @param o the <code>double</code> that should be visited
  */
 public void visit(double o)  { visitDefault(o); }
 /**
  * A "catch-all" <code>visit</code> method, redirected to
  * <code>visitDefault</code>.
  * @param o the <code>short</code> that should be visited
  */
 public void visit(short o)   { visitDefault(o); }
 /**
  * A "catch-all" <code>visit</code> method, redirected to
  * <code>visitDefault</code>.
  * @param o the <code>boolean</code> that should be visited
  */
 public void visit(boolean o) { visitDefault(o); }

 /**
  * The bridge method that performs the actual
  * dispatching, calling the appropriate
  * <code>visit</code> for this object.
  * @param o the object that should be visited
  */
 public void visitAppropriate(Object o)  { visitAppropriate(o,o.getClass()); }
 /**
  * The bridge method that performs the dispatching
  * for objects or primitives.
  * @param o the object that should be visited
  *  (possibly a wrapped primitive)
  * @param c the class of the object (possibly
  *  a primitive type)
  */
 public void visitAppropriate(Object o,Class c)  { 
  if (c==char.class)
   theDispatcher.visit(((Character)o).charValue());
  else if (c==byte.class)
   theDispatcher.visit(((Byte)o).byteValue());
  else if (c==int.class)
   theDispatcher.visit(((Integer)o).intValue());
  else if (c==long.class)
   theDispatcher.visit(((Long)o).longValue());
  else if (c==float.class)
   theDispatcher.visit(((Float)o).floatValue());
  else if (c==double.class)
   theDispatcher.visit(((Double)o).doubleValue());
  else if (c==short.class)
   theDispatcher.visit(((Short)o).shortValue());
  else if (c==boolean.class)
   theDispatcher.visit(((Boolean)o).booleanValue());
  else
   theDispatcher.visit(o);
 }
 
 final private RunaboutBase theDispatcher;
 final private static AbstractCollection<Class> registered;
 static {
  registered=new java.util.HashSet<Class>();
 }
 protected RunaboutBis() {
  synchronized (registered) {
   if (!(registered.contains(this.getClass()))) {
    registered.add(this.getClass());
    ovm.polyd.Descriptor<RunaboutBase> d2;
    d2 = new ovm.polyd.Descriptor<RunaboutBase>(RunaboutBase.class,new Class[] {this.getClass()});
    d2.setDispatching(RunaboutDisp.class);
    d2.register();
   }
  }
  theDispatcher = (RunaboutBase)PolyD.buildFromDescriptor(RunaboutBase.class, new Object[] {this});
  // System.out.println("Creating new "+this.getClass());
 }
}
