/*
  (c) 2005 Antonio Cunei, All rights reserved.

  This file is part of the Runabout Emulation Layer for PolyD.

  The Runabout Emulation Layer for PolyD is free software;
  you can redistribute it and/or modify it under the terms of
  the GNU General Public License as published by the Free
  Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  The Runabout Emulation Layer for PolyD is distributed
  in the hope that it will be useful, but WITHOUT ANY WARRANTY;
  without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE. See the GNU General Public License
  for more details.

  You should have received a copy of the GNU General Public License
  along with the Runabout Emulation Layer for PolyD; if not, write
  to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
  Boston, MA  02111-1307  USA
*/

/*
 *   The original Runabout is (C) Christian Grothoff and it is
 *   released under the GPL, GNU Public License. This file
 *   is a derivative work of the Runabout, and it is also
 *   released under the terms of the GNU Public License.
 */

package ovm.polyd.runemu;

import java.io.PrintStream;
import java.lang.reflect.Method;
import ovm.polyd.policy.Dispatching;
import ovm.polyd.policy.MultiDisp;
import ovm.polyd.Utils;

/**
 *   A dispatching policy that reproduces the method selection policy
 *   used by the original Runabout. It is derived from
 *   parts of the original Runabout source code, and
 *   it is therefore released under the terms of the GNU Public License.
 * 
 * @version $Revision: 1.2 $, $Date: 2005/04/02 21:51:37 $
 * @author Antonio Cunei
 *
 */
public class RunaboutDisp extends MultiDisp
{

    private static final Dispatching theDispatcher = new RunaboutDisp();

    /**
     * Returns an instance of this dispatching policy.
     * 
     * @return an instance of this policy (a singleton)
     */
    public static Dispatching theDispatcher()
    {
        return theDispatcher;
    }

    public int bestMatch(Class args[], Method meth[])
    {
        if(meth.length == 0)
            return -1;

        // we only deal with unary methods ("visit") here
        
        if(args.length != 1)
         return -1;

        // we need to reproduce somehow the selection of the best method, as
        // applied by the Runabout. That implies giving preference to superclasses,
        // and the inspecting the interfaces.
        // We know that selection has to be applied in the list of methods
        // supplied. We also know that this dispatcher will only ever be called by
        // the runabout emulation layer. Therefore the method list is actually equivalent
        // to a list of Classes, from which I have to select.
        
        Class[] cc=new Class[meth.length];
        for (int i=0;i<meth.length;i++)
         cc[i]=meth[i].getParameterTypes()[0];
        
        // Now we know that we have to pick our candidate in "cc"
        
        Class c=args[0];
        
        // ok, let's find the best match for arg within candidates, following
        // the Runabout approach.

        // Note: in the original Runabout there is no "visit(Object)"
        // method within the Runabout class. Therefore such a method
        // is never found. But we need to have a visit(Object)
        // in the emulated Runabout class in order to satisfy the
        // interface. However, such a method would always be found
        // scanning the superclasses, and the interfaces would be
        // ignored. So, instead, we ignore while scanning, and we
        // resort to visit(Object) only as a desperate measure.

        int pos=findInArray(c,cc);
        if (pos>=0) return pos; //we're done!
        Class cl = c.getSuperclass();
        while (cl != null) {
         int co = findNotObjectInArray(cl, cc);
         if (co >= 0)
          return co;
         cl = cl.getSuperclass();
        }
        int result=scanInterfaces(c, c, cc);
        if (result>=0) return result;
        // couldn't find. Return visit(Object)
        return findInArray(Object.class,cc);
    }

 private int findNotObjectInArray(Class c, Class[] cc) {
  if (c==Object.class) return -1;
  return findInArray(c,cc);
 }

 private int findInArray(Class c, Class[] cc) {
  // bleah
  int d=cc.length;
  for (int i=0;i<d;i++)
   if (c==cc[i]) return i;
  return -1;
 }

 private int scanInterfaces(Class c, Class cl, Class[] cc) {
  int co = -1;
  while (cl != null) {
   Class[] ifc = cl.getInterfaces();
   for (int i = 0; i < ifc.length; i++) {
    int r = findNotObjectInArray(ifc[i], cc);
    if (r >= 0) {
     if ((co >= 0) && (r != co)) {
      throw new RunaboutException(
       "Ambiguous resolution for visit call to "+c+" in "+getClass().getName()
      );
     }
     co = r;
    }
   }
   for (int i = 0; i < ifc.length; i++) {
    int r = scanInterfaces(c, ifc[i], cc);
    if (r >= 0) {
     if ((co >= 0) && (r != co)) {
      throw new RunaboutException(
       "Ambiguous resolution for visit call to "+c+" in "+getClass().getName()
      );
     }
     co = r;
    }
   }
   cl = cl.getSuperclass();
  }
  return co;
 }
 public String toString() {
  return "RunaboutDisp";
 }

}
