/*
  (c) 2005 Antonio Cunei, All rights reserved.

  This file is part of the Runabout Emulation Layer for PolyD.

  The Runabout Emulation Layer for PolyD is free software;
  you can redistribute it and/or modify it under the terms of
  the GNU General Public License as published by the Free
  Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  The Runabout Emulation Layer for PolyD is distributed
  in the hope that it will be useful, but WITHOUT ANY WARRANTY;
  without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE. See the GNU General Public License
  for more details.

  You should have received a copy of the GNU General Public License
  along with the Runabout Emulation Layer for PolyD; if not, write
  to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
  Boston, MA  02111-1307  USA
*/

package ovm.polyd.runemu;

import ovm.polyd.policy.*;
import ovm.polyd.PolyD;
/**
 * A Runabout emulation that will gather statistics while it runs.<p>
 * Use <code>RunaboutStat</code> as you would use a conventional Runabout.
 * The method <code>printStats</code> can be used at any time 
 * to print out a summary of the dispatchers created using the
 * <code>RunaboutStat</code> as a base class, and of the messages dispatched
 * using those dispatchers. This emulation does not support primitives.
 * 
 * @version $Revision: 1.2 $, $Date: 2005/04/02 21:51:37 $
 * @author Antonio Cunei
 *
 */
public class RunaboutStat {

 protected void err(Class c) {
  throw new RunaboutException(
   "The Runabout compatibility layer of PolyD has detected\n"+
   "a visit() call without a proper handler,\n"+
   "when dispatching using "+this.getClass()+",\n"+
   "calling visit("+c.getName()+")"
  );
 }
 /**
  * A "catch-all" <code>visit</code> method, redirected to
  * <code>visitDefault</code>.
  * @param o the object that should be visited
  */
 public void visit(Object o)  { visitDefault(o); }
 // Can only be called in the original Runabout using
 // visitAppropriate(Object o,Class c), which is *not* implemented here
 // If you need primitives, please use RunaboutBis, in this package
 /**
  * Unused
  */
 public void visit(char o)    { }
 /**
  * Unused
  */
 public void visit(byte o)    { }
 /**
  * Unused
  */
 public void visit(int o)     { }
 /**
  * Unused
  */
 public void visit(long o)    { }
 /**
  * Unused
  */
 public void visit(float o)   { }
 /**
  * Unused
  */
 public void visit(double o)  { }
 /**
  * Unused
  */
 public void visit(short o)   { }
 /**
  * Unused
  */
 public void visit(boolean o) { }

 protected void visitDefault(Object o)  { err(o.getClass()); }

 /**
  * The bridge method that performs the actual
  * dispatching, calling the appropriate
  * <code>visit</code> for this object.
  * @param obj the object that should be visited
  */
 public void visitAppropriate(Object obj)  { 
  theDispatcher.visit(obj);
  }
 
 /**
  * The bridge method that performs the dispatching
  * for objects or primitives (not available in this emulation).
  * @param o the object that should be visited
  *  (possibly a wrapped primitive)
  * @param c the class of the object (possibly
  *  a primitive type)
  */
 public void visitAppropriate(Object o,Class c) {
  throw new RunaboutException(
    "RunaboutStat does not handle primitives.\n"+
    "If you need primitives, please use ovm.polyd.runemu.RunaboutBis,\n"+
    "if you only need objects, please use visitAppropriate(Object) instead.");
 }
 
 final private RunaboutBase theDispatcher;
 
 private static java.util.Map stats=new java.util.HashMap();

 static private class Counter {
  public int i;
 }

 protected RunaboutStat() {
  Counter ii;
  synchronized (Runabout.class) {
   if (stats.containsKey(this.getClass()))
    ii=(Counter)(stats.get(this.getClass()));
   else {
    ii=new Counter();
    // ii.i=0; // default
    stats.put(this.getClass(),ii);
    ovm.polyd.Descriptor d2;
    d2 = new ovm.polyd.Descriptor(RunaboutBase.class,new Class[] {this.getClass()});
    d2.setDispatching(RunaboutDisp.class);
    // d2.setInvocation(PlainInvocation.class);
    d2.setInvocation(RunaboutStatInvocation.class);
    d2.register();
   }
   ii.i++;
  }
  theDispatcher = (RunaboutBase)PolyD.buildFromDescriptor(RunaboutBase.class, new Object[] {this});
  // System.out.println("Creating new "+this.getClass());
 }
 
/**
 * Prints a summary of dispatchers and the messages
 * that they have dispatched so far.
 */
 public static void printStats() {
  synchronized (Runabout.class) {
   java.util.Iterator ll=stats.keySet().iterator();
   int t=0;
   while (ll.hasNext()) {
    Class c=(Class)(ll.next());
    Counter ii=(Counter)(stats.get(c));
    System.out.println("Creations:   "+c.getName()+" created "+ii.i+" times");
    t+=ii.i;
   }
   System.out.println("Total: "+t+" Runabouts created.");
  }
  System.out.println("---------------");
  RunaboutStatInvocation.printStats();
 }
}
